@@ -81,13 +81,14 @@ static void server_options(char **args,int *argc)
   int ac = *argc;
   static char argstr[50];
   static char bsize[30];
-  int i, x = 1;
+  int i, x;
 
   args[ac++] = "--server";
 
   if (!sender)
     args[ac++] = "--sender";
 
+  x = 1;
   argstr[0] = '-';
   for (i=0;i<verbose;i++)
     argstr[x++] = 'v';
@@ -117,7 +118,7 @@ static void server_options(char **args,int *argc)
     argstr[x++] = 'C';
   argstr[x] = 0;
 
-  args[ac++] = argstr;
+  if (x != 1) args[ac++] = argstr;
 
   if (block_size != BLOCK_SIZE) {
     sprintf(bsize,"-B%d",block_size);
@@ -204,7 +205,7 @@ static char *get_local_name(struct file_list *flist,char *name)
       return NULL;
     }
     if (flist->count > 1) {
-      fprintf(stderr,"ERROR: destination must be a diirectory when copying more than 1 file\n");
+      fprintf(stderr,"ERROR: destination must be a directory when copying more than 1 file\n");
       exit(1);
     }
     return name;
@@ -213,6 +214,9 @@ static char *get_local_name(struct file_list *flist,char *name)
   if (flist->count == 1)
     return name;
 
+  if (!name) 
+    return NULL;
+
   if (mkdir(name,0777) != 0) {
     fprintf(stderr,"mkdir %s : %s\n",name,strerror(errno));
     exit(1);
@@ -280,6 +284,10 @@ void do_server_recv(int argc,char *argv[])
     dir = argv[0];
     argc--;
     argv++;
+    if (chdir(dir) != 0) {
+      fprintf(stderr,"chdir %s : %s\n",dir,strerror(errno));
+      exit(1);
+    }    
   }
 
   flist = recv_file_list(STDIN_FILENO);
@@ -288,12 +296,17 @@ void do_server_recv(int argc,char *argv[])
     exit(1);
   }
 
-
-  local_name = get_local_name(flist,argv[0]);
+  if (argc > 0) {    
+    if (strcmp(dir,".")) {
+      argv[0] += strlen(dir);
+      if (argv[0][0] == '/') argv[0]++;
+    }
+    local_name = get_local_name(flist,argv[0]);
+  }
 
   if ((pid=fork()) == 0) {
     recv_files(STDIN_FILENO,flist,local_name);
-    if (verbose > 1)
+    if (verbose > 2)
       fprintf(stderr,"receiver read %d\n",read_total());
     exit(0);
   }
@@ -387,7 +400,7 @@ int main(int argc,char *argv[])
 
     while ((opt = getopt_long(argc, argv, 
 			      short_options, long_options, &option_index)) 
-	   != -1)
+	   != -1) {
       switch (opt) 
 	{
 	case OPT_VERSION:
@@ -404,7 +417,7 @@ int main(int argc,char *argv[])
 	  break;
 
 	case OPT_EXCLUDE_FROM:
-	  add_exclude_file(optarg);
+	  add_exclude_file(optarg,1);
 	  break;
 
 	case 'h':
@@ -513,6 +526,7 @@ int main(int argc,char *argv[])
 	  fprintf(stderr,"bad option -%c\n",opt);
 	  exit(1);
 	}
+    }
 
     while (optind--) {
       argc--;
