@@ -20,6 +20,7 @@
 #include "rsync.h"
 
 extern int verbose;
+extern int am_server;
 
 typedef unsigned short tag;
 
@@ -29,6 +30,13 @@ typedef unsigned short tag;
 static int false_alarms;
 static int tag_hits;
 static int matches;
+static int data_transfer;
+
+static int total_false_alarms=0;
+static int total_tag_hits=0;
+static int total_matches=0;
+static int total_data_transfer=0;
+
 
 struct target {
   tag t;
@@ -39,13 +47,15 @@ static struct target *targets=NULL;
 
 static tag tag_table[TABLESIZE];
 
-#define gettag(sum) (((sum)>>16) + ((sum)&0xFFFF))
+#define gettag2(s1,s2) (((s1) + (s2)) & 0xFFFF)
+#define gettag(sum) gettag2((sum)&0xFFFF,(sum)>>16)
 
 static int compare_targets(struct target *t1,struct target *t2)
 {
   return(t1->t - t2->t);
 }
 
+
 static void build_hash_table(struct sum_struct *s)
 {
   int i;
@@ -85,6 +95,7 @@ static void matched(int f,struct sum_struct *s,char *buf,off_t len,int offset,in
   if (n > 0) {
     write_int(f,n);
     write_buf(f,buf+last_match,n);
+    data_transfer += n;
   }
   write_int(f,-(i+1));
   if (i != -1)
@@ -106,7 +117,7 @@ static void hash_search(int f,struct sum_struct *s,char *buf,off_t len)
 
   k = MIN(len, s->n);
   sum = get_checksum1(buf, k);
-  s1 = sum;
+  s1 = sum & 0xFFFF;
   s2 = sum >> 16;
   if (verbose > 3)
     fprintf(stderr, "sum=%.8x k=%d\n", sum, k);
@@ -120,7 +131,7 @@ static void hash_search(int f,struct sum_struct *s,char *buf,off_t len)
 	    s->n,(int)len,s->count);
 
   do {
-    tag t = (s1 + s2) & 0xffff;		/* gettag(sum) */
+    tag t = gettag2(s1,s2);
     j = tag_table[t];
     if (verbose > 4)
       fprintf(stderr,"offset=%d sum=%08x\n",
@@ -129,7 +140,7 @@ static void hash_search(int f,struct sum_struct *s,char *buf,off_t len)
     if (j != NULL_TAG) {
       int done_csum2 = 0;
 
-      sum = (s1 & 0xffff) + (s2 << 16);
+      sum = (s1 & 0xffff) | (s2 << 16);
       tag_hits++;
       do {
 	int i = targets[j].i;
@@ -148,7 +159,7 @@ static void hash_search(int f,struct sum_struct *s,char *buf,off_t len)
 	    offset += s->sums[i].len - 1;
 	    k = MIN((len-offset), s->n);
 	    sum = get_checksum1(buf+offset, k);
-	    s1 = sum;
+	    s1 = sum & 0xFFFF;
 	    s2 = sum >> 16;
 	    ++matches;
 	    break;
@@ -187,6 +198,8 @@ void match_sums(int f,struct sum_struct *s,char *buf,off_t len)
   last_match = 0;
   false_alarms = 0;
   tag_hits = 0;
+  matches=0;
+  data_transfer=0;
 
   if (len > 0 && s->count>0) {
     build_hash_table(s);
@@ -210,4 +223,20 @@ void match_sums(int f,struct sum_struct *s,char *buf,off_t len)
   if (verbose > 2)
     fprintf(stderr, "false_alarms=%d tag_hits=%d matches=%d\n",
 	    false_alarms, tag_hits, matches);
+
+  total_tag_hits += tag_hits;
+  total_false_alarms += false_alarms;
+  total_matches += matches;
+  total_data_transfer += data_transfer;
+}
+
+void match_report(void)
+{
+  if (verbose <= 1)
+    return;
+
+  fprintf(am_server?stderr:stdout,
+	  "total: matches=%d  tag_hits=%d  false_alarms=%d  data=%d\n",
+	  total_matches,total_tag_hits,
+	  total_false_alarms,total_data_transfer);
 }
