@@ -183,38 +183,66 @@ static struct sum_struct *receive_sums(int f)
 }
 
 
-static void set_perms(char *fname,struct file_struct *file)
+static void set_perms(char *fname,struct file_struct *file,struct stat *st,
+		      int report)
 {
+  int updated = 0;
+  struct stat st2;
+
   if (dry_run) return;
 
+  if (!st) {
+    if (stat(fname,&st2) != 0) {
+      fprintf(stderr,"stat %s : %s\n",fname,strerror(errno));
+      return;
+    }
+    st = &st2;
+  }
+
 #ifdef HAVE_UTIME
-  if (preserve_times) {
+  if (preserve_times && st->st_mtime != file->modtime) {
     struct utimbuf tbuf;  
+    updated = 1;
     tbuf.actime = time(NULL);
     tbuf.modtime = file->modtime;
-    if (utime(fname,&tbuf) != 0)
+    if (utime(fname,&tbuf) != 0) {
       fprintf(stderr,"failed to set times on %s : %s\n",
 	      fname,strerror(errno));
+      return;
+    }
   }
 #endif
 
 #ifdef HAVE_CHMOD
-  if (preserve_perms) {
-    if (chmod(fname,file->mode) != 0)
+  if (preserve_perms && st->st_mode != file->mode) {
+    updated = 1;
+    if (chmod(fname,file->mode) != 0) {
       fprintf(stderr,"failed to set permissions on %s : %s\n",
 	      fname,strerror(errno));
+      return;
+    }
   }
 #endif
 
 #ifdef HAVE_CHOWN
-  if (preserve_uid || preserve_gid) {
+  if ((preserve_uid && st->st_uid != file->uid) || 
+      (preserve_gid && st->st_gid != file->gid) {
+    updated = 1;
     if (chown(fname,
 	      preserve_uid?file->uid:-1,
 	      preserve_gid?file->gid:-1) != 0) {
       fprintf(stderr,"chown %s : %s\n",fname,strerror(errno));
+      return;
     }
   }
 #endif
+
+  if (verbose > 1 && report) {
+    if (updated)
+      fprintf(am_server?stderr:stdout,"%s\n",fname);
+    else
+      fprintf(am_server?stderr:stdout,"%s is uptodate\n",fname);
+  }
 }
 
 
@@ -230,7 +258,7 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
   if (verbose > 2)
     fprintf(stderr,"recv_generator(%s)\n",fname);
 
-  statret = stat(fname,&st);
+  statret = lstat(fname,&st);
 
 #if SUPPORT_LINKS
   if (preserve_links && S_ISLNK(flist->files[i].mode)) {
@@ -240,8 +268,11 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
       l = readlink(fname,lnk,MAXPATHLEN-1);
       if (l > 0) {
 	lnk[l] = 0;
-	if (strcmp(lnk,flist->files[i].link) == 0)
+	if (strcmp(lnk,flist->files[i].link) == 0) {
+	  if (verbose > 1) 
+	    fprintf(am_server?stderr:stdout,"%s is uptodate\n",fname);
 	  return;
+	}
       }
     }
     if (!dry_run) unlink(fname);
@@ -266,13 +297,16 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
       if (verbose > 2)
 	fprintf(stderr,"mknod(%s,0%o,0x%x)\n",
 		fname,(int)flist->files[i].mode,(int)flist->files[i].dev);
-      if (!dry_run && mknod(fname,flist->files[i].mode,flist->files[i].dev) != 0) {
+      if (!dry_run && 
+	  mknod(fname,flist->files[i].mode,flist->files[i].dev) != 0) {
 	fprintf(stderr,"mknod %s : %s\n",fname,strerror(errno));
       } else {
+	set_perms(fname,&flist->files[i],NULL,0);
 	if (verbose)
 	  fprintf(am_server?stderr:stdout,"%s\n",fname);
-	set_perms(fname,&flist->files[i]);
       }
+    } else {
+      set_perms(fname,&flist->files[i],&st,1);
     }
     return;
   }
@@ -309,12 +343,11 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
     file_checksum(fname,sum,st.st_size);
   }
 
-  if ((st.st_size == flist->files[i].length &&
-       ((!preserve_perms || st.st_mtime == flist->files[i].modtime) ||
-	(S_ISREG(st.st_mode) && 
-	 always_checksum && memcmp(sum,flist->files[i].sum,SUM_LENGTH) == 0)))) {
-    if (verbose > 1)
-      fprintf(stderr,"%s is uptodate\n",fname);
+  if (st.st_size == flist->files[i].length &&
+      (st.st_mtime == flist->files[i].modtime ||
+       (always_checksum && S_ISREG(st.st_mode) && 	  
+	memcmp(sum,flist->files[i].sum,SUM_LENGTH) == 0))) {
+    set_perms(fname,&flist->files[i],&st,1);
     return;
   }
 
@@ -345,7 +378,7 @@ void recv_generator(char *fname,struct file_list *flist,int i,int f_out)
   if (verbose > 3)
     fprintf(stderr,"mapped %s of size %d\n",fname,(int)st.st_size);
 
-  s = generate_sums(buf,st.st_size,BLOCK_SIZE);
+  s = generate_sums(buf,st.st_size,block_size);
 
   write_int(f_out,i);
   send_sums(s,f_out);
@@ -496,12 +529,13 @@ int recv_files(int f_in,struct file_list *flist,char *local_name)
 
       /* move tmp file over real file */
       if (rename(fnametmp,fname) != 0) {
-	fprintf(stderr,"rename %s -> %s : %s\n",fnametmp,fname,strerror(errno));
+	fprintf(stderr,"rename %s -> %s : %s\n",
+		fnametmp,fname,strerror(errno));
       }
 
       unmap_file(buf,st.st_size);
 
-      set_perms(fname,&flist->files[i]);
+      set_perms(fname,&flist->files[i],NULL,0);
     }
 
   if (verbose > 2)
@@ -547,17 +581,17 @@ off_t send_files(struct file_list *flist,int f_out,int f_in)
 	continue;
       }
 
-      fd = open(fname,O_RDONLY);
-      if (fd == -1) {
-	fprintf(stderr,"send_files failed to open %s: %s\n",
-		fname,strerror(errno));
-	return -1;
-      }
-  
       s = receive_sums(f_in);
       if (!s) 
 	return -1;
 
+      fd = open(fname,O_RDONLY);
+      if (fd == -1) {
+	fprintf(stderr,"send_files failed to open %s: %s\n",
+		fname,strerror(errno));
+	continue;
+      }
+  
       /* map the local file */
       if (fstat(fd,&st) != 0) 
 	return -1;
@@ -599,6 +633,8 @@ off_t send_files(struct file_list *flist,int f_out,int f_in)
       total += st.st_size;
     }
 
+  match_report();
+
   write_int(f_out,-1);
   write_flush(f_out);
 
@@ -616,6 +652,7 @@ void generate_files(int f,struct file_list *flist,char *local_name)
 	    (int)getpid(),flist->count);
 
   for (i = 0; i < flist->count; i++) {
+    if (!flist->files[i].name) continue;
     if (S_ISDIR(flist->files[i].mode)) {
       if (dry_run) continue;
       if (mkdir(flist->files[i].name,flist->files[i].mode) != 0 &&
@@ -630,6 +667,6 @@ void generate_files(int f,struct file_list *flist,char *local_name)
   }
   write_int(f,-1);
   write_flush(f);
-  if (verbose > 1)
+  if (verbose > 2)
     fprintf(stderr,"generator wrote %d\n",write_total());
 }
